#!/bin/bash

usage="cube-root-p2.bash <target> [digits]"

export BC_LINE_LENGTH=200

p=2

target=$1

digits=$2

shopt -s extglob

if [[ $target != *([0-9])[13579] ]]
then
    echo "error: target must be positive and of the form 2n+1"
    exit 1
fi

if [[ -n $digits ]]
then
    if [[ $digits != +([0-9]) || $digits -gt $BC_LINE_LENGTH ]]
    then
	echo "error: digits must be less than $BC_LINE_LENGTH"
	exit 1
    fi
else
    digits=25
fi

cbrt=1

while :
do
  cbrt_base_2=$(echo -e "obase=2 \n $cbrt" | bc -q)

  approx=$(echo -e "obase=2 \n $cbrt*$cbrt*$cbrt" | bc -q)

  printf "%${digits}s %$((digits * 3))s\n" $cbrt_base_2 $approx

  result=$(echo -e "obase=2 \n $cbrt*$cbrt*$cbrt - $target" | bc -q)

  zeros=$(echo "$result" | awk '{s=length($0); sub("0+$",""); t=length($0); print s-t;}')

  [[ $zeros -ge $digits ]] && break

  incr=$(echo "2 ^ $zeros" | bc -q)

  cbrt=$((cbrt+incr))
done

echo "cube root ($target) ~ $cbrt"
