#!/bin/bash

# This computes the inverse of a unit in the ring of integers mod p to
# some power.  The first argument is p - the prime number which
# partially determines the size of the ring.  The second argument is
# the unit to invert.  The third argument is the power of p that
# determines the number of elements in the ring.

# The technique used to compute the inverse is successive
# approximation.  At the beginning of the main "while" loop of the
# algorithm, assume that we know the inverse of x accurate to i
# base-p digits.  Each iteration through the "while" loop determines
# one additional base-p digit of accuracy in the inverse.

# The approximation computes the new digit by multiplying the current
# approximation by x and then determining the digit of that product in
# the i-th location.  The new digit is chosen so that when it is added
# to the current inverse, the i-th digit of the product will be made
# equal to 0.

export BC_LINE_LENGTH=200

p=$1

x=$2

digits=$3

seed=$(echo -e "$x % $p" | bc -q)

if [[ $seed == 0 ]]
then
    echo "error: $x is not a unit in this ring"
    exit 1
fi

declare -i j=1

while (( (seed*j)%p != 1 ))
do
  j=j+1
done

seed=$j

xinv=0
xinv_base_p=0

declare -i i=0

printf "%$((digits-1))s%0${i}s %s\n" "" $xinv_base_p $xinv

while [[ $i -lt $digits ]]
do
  d=$(echo -e "( $p + ($x * $xinv - 1) / ($p ^ $i) ) % $p" | bc -q)

  d=$(echo -e "( ($p - $d)* $seed) % $p" | bc -q)

  xinv=$(echo -e "$xinv + $d * ($p ^ $i)" | bc -q)

  xinv_base_p=$(echo -e "obase=$p \n $xinv" | bc -q)

  i=i+1

  printf "%$((digits-i))s%0${i}s %s\n" "" $xinv_base_p $xinv
done
