#!/bin/bash

usage="sqrt-tot.bash -p <prime> -t <target> [-a|-s <source-digit>] [-d <digits>]"

shopt -s extglob

while getopts "p:t:s:d:a" c
do
  case $c in
      p) p=$OPTARG;;
      t) target=$OPTARG;;
      a) alternate=1;;
      s) source=$OPTARG;;
      d) digits=$OPTARG;;
      *) echo "$usage"
         exit 1;;
  esac
done

if [[ -z $p || -z $target || (-n $alternate && -n $source) ]]
then
    echo "$usage"
    exit 1
fi

if [[ $p == 2 ]]
then
    echo "error: -p argument $p must be a prime different from 2"
    exit 1
fi

prime_check=$(factor $p)

if [[ $prime_check != @($p: $p) ]]
then
    echo "error: -p argument $p must be prime"
    exit 1
fi

if [[ $target != ?(-)+([0-9]) ]]
then
    echo "error: -t argument $d must be a digit string"
    exit 1
fi

if [[ -n $source && ( $source != ?(-)+([0-9]) || $source -lt 0 || $source -gt $p ) ]]
then
    echo "error: -s argument $s must be a digit string in the range [0,$p)"
    exit 1
fi

if [[ -n $digits && $digits != +([0-9]) ]]
then
    echo "error: -d argument $d must be a positive number"
    exit 1
fi

if [[ -z $digits ]]
then
    [[ $p -lt 16 ]] && digits=20
    [[ $p -gt 16 ]] && digits=10
fi

export BC_LINE_LENGTH=200

################ check for final digit ################

v1=$target

while [[ $v1 -lt 0 ]]
do
  v1=$(( v1 + p ))
done

v1=$(( v1 % p ))

if [[ -n $source ]]
then
    v2=$(( ( source * source ) % p ))

    if [[ $v1 != $v2 ]]
    then
	echo "error: inconsistent choice of <target> and <source>"
	exit 1
    fi
else
    i=1
    while [[ $i -lt $p ]]
    do
      v2=$(( ( i * i ) % p ))

      [[ $v1 == $v2 ]] && break

      i=$((i+1))
    done

    if [[ $v1 != $v2 ]]
    then
	echo "error: no initial digit exists for square root"
	exit 1
    fi

    if [[ -z $alternate ]]
    then
	source=$i
    else
	source=$(( (i * (p-1)) % p ))
    fi
fi


################ initialize lookup table ################

sqrt=$source

i=0
while [[ $i -lt $p ]]
do
  j=$(( (2 * sqrt * i) % p ))
  j=$(( (p - j) % p ))
  digit_lookup[j]=$i
  i=$((i+1))
done

################ begin iteration ################

if [[ $p -lt 16 ]]
then
    format_digits=$digits
else
    format_digits=$((digits * 3))
fi

position=1

while :
do
  [[ $position -eq $digits ]] && break

  sqrt_base_p=$(echo -e "obase=$p \n $sqrt" | bc -q)

  sqr_base_p=$(echo -e "obase=$p \n $sqrt * $sqrt" | bc -q)

  printf "%${format_digits}s %$((format_digits * 2))s\n" "$sqrt_base_p" "$sqr_base_p"

  offset=0

  while :
  do
    d=$(echo "(($sqrt*$sqrt - $target + $offset * ($p^($position+1))) % ($p^($position+1)))/($p^$position) " | bc -q)

    [[ $d -ge 0 ]] && break

    offset=$((offset+1))
  done

  new_digit=${digit_lookup[d]}

  sqrt=$(echo "$sqrt + ($new_digit * ($p ^ $position))" | bc -q)

  position=$((position + 1))
done

echo "sqrt($target) ~ $sqrt"
